@@ -1,5 +1,7 @@
+use std::f32::INFINITY;
+
 use cpython::{exc, ObjectProtocol, PyDict, PyErr, PyObject, PyResult, Python};
-use dubins_paths::{path_sample, shortest_path, DubinsError, DubinsPath};
+use dubins_paths::{intermediate_results, path_sample, word, DubinsError, DubinsPath, DubinsPathType};
 use rl_ball_sym::simulation::ball::Ball;
 
 use vvec3::Vec3;
@@ -32,6 +34,7 @@ pub const END_THROTTLE_ACCEL_B: f32 = 160.;
 const BOOST_ACCEL: f32 = 991. + 2. / 3.;
 const BOOST_ACCEL_DT: f32 = BOOST_ACCEL * SIMULATION_DT;
 
+#[derive(Clone)]
 pub struct Hitbox {
     pub length: f32,
     pub width: f32,
@@ -58,15 +61,18 @@ impl Hitbox {
     }
 }
 
+#[derive(Clone)]
 pub struct Car {
     pub location: Vec3,
     pub velocity: Vec3,
+    pub local_velocity: Vec3,
     pub angular_velocity: Vec3,
     pub forward: Vec3,
     pub right: Vec3,
     pub up: Vec3,
     pub hitbox: Hitbox,
     pub hitbox_offset: Vec3,
+    pub field: CarFieldRect,
     pub pitch: f32,
     pub yaw: f32,
     pub roll: f32,
@@ -85,12 +91,14 @@ impl Default for Car {
         Self {
             location: Vec3::default(),
             velocity: Vec3::default(),
+            local_velocity: Vec3::default(),
             angular_velocity: Vec3::default(),
             forward: Vec3::default(),
             right: Vec3::default(),
             up: Vec3::default(),
             hitbox: Hitbox::default(),
             hitbox_offset: Vec3::default(),
+            field: CarFieldRect::default(),
             pitch: 0.,
             yaw: 0.,
             roll: 0.,
@@ -114,14 +122,23 @@ impl Car {
         let c_r = self.roll.cos();
         let s_r = self.roll.sin();
 
+        // CP * CY
+        // CP * SY
+        // SP
         self.forward.x = c_p * c_y;
         self.forward.y = c_p * s_y;
         self.forward.z = s_p;
 
+        // CY * SP * SR - CR * SY
+        // SY * SP * SR + CR * CY
+        // -CP * SR
         self.right.x = c_y * s_p * s_r - c_r * s_y;
         self.right.y = s_y * s_p * s_r + c_r * c_y;
-        self.right.z = -c_p * c_r;
+        self.right.z = -c_p * s_r;
 
+        // -CR * CY * SP - SR * SY
+        // -CR * SY * SP + SR * CY
+        // CP * CR
         self.up.x = -c_r * c_y * s_p - s_r * s_y;
         self.up.y = -c_r * s_y * s_p + s_r * c_y;
         self.up.z = c_p * c_r;
@@ -130,8 +147,14 @@ impl Car {
     pub fn calculate_max_values(&mut self) {
         self.max_speed = self.get_max_speed();
         self.msmtr = turn_radius(self.max_speed);
-        // self.max_speed = MAX_SPEED;
-        // self.msmtr = MAX_SPEED_MIN_TURN_RADIUS;
+    }
+
+    pub fn calculate_local_values(&mut self) {
+        self.local_velocity = localize(self, self.velocity);
+    }
+
+    pub fn calculate_field(&mut self) {
+        self.field = CarFieldRect::from(&self.hitbox);
     }
 
     fn get_max_speed(&mut self) -> f32 {
@@ -151,14 +174,13 @@ impl Car {
                 }
             }
 
-            if v.signum() == 1. {
-                v += throttle_acceleration(v) * SIMULATION_DT;
+            if v.signum() == -1. {
+                v += BRAKE_ACC_DT;
             } else {
-                v += -BRAKE_ACC_DT;
+                v += throttle_acceleration(v) * SIMULATION_DT;
+                v += BOOST_ACCEL_DT;
+                b -= BOOST_CONSUMPTION_DT;
             }
-
-            v += BOOST_ACCEL_DT;
-            b -= BOOST_CONSUMPTION_DT;
         }
     }
 }
@@ -209,6 +231,17 @@ pub fn get_f32_from_dict(py: Python, py_dict: &PyDict, key: &str, name: &str) ->
     }
 }
 
+pub fn get_usize_from_dict(py: Python, py_dict: &PyDict, key: &str, name: &str) -> PyResult<usize> {
+    match py_dict.get_item(py, key) {
+        Some(py_num) => {
+            return Ok(py_num.extract(py)?);
+        }
+        None => {
+            return Err(PyErr::new::<exc::AttributeError, _>(py, format!("No key called '{}' in '{}'.", key, name)));
+        }
+    }
+}
+
 pub fn get_u8_from_dict(py: Python, py_dict: &PyDict, key: &str, name: &str) -> PyResult<u8> {
     match py_dict.get_item(py, key) {
         Some(py_num) => {
@@ -235,9 +268,11 @@ pub fn get_vec_from_vec3(vec: Vec3) -> Vec<f32> {
     vec![vec.x, vec.y, vec.z]
 }
 
-pub fn localize_2d(car: &Car, vec: Vec3) -> Vec3 {
-    let vec = vec - car.location;
+pub fn localize_2d_location(car: &Car, vec: Vec3) -> Vec3 {
+    localize_2d(car, vec - car.location)
+}
 
+pub fn localize_2d(car: &Car, vec: Vec3) -> Vec3 {
     Vec3 {
         x: vec.dot(&car.forward),
         y: vec.dot(&car.right),
@@ -245,16 +280,18 @@ pub fn localize_2d(car: &Car, vec: Vec3) -> Vec3 {
     }
 }
 
-// pub fn localize(car: &Car, vec: Vec3) -> Vec3 {
-//     let vec = vec - car.location;
-
-//     Vec3 {
-//         x: vec.dot(&car.forward),
-//         y: vec.dot(&car.right),
-//         z: vec.dot(&car.up),
-//     }
+// pub fn localize_location(car: &Car, vec: Vec3) -> Vec3 {
+//     localize(car, vec - car.location)
 // }
 
+pub fn localize(car: &Car, vec: Vec3) -> Vec3 {
+    Vec3 {
+        x: vec.dot(&car.forward),
+        y: vec.dot(&car.right),
+        z: vec.dot(&car.up),
+    }
+}
+
 // pub fn globalize(car: &Car, vec: Vec3) -> Vec3 {
 //     car.forward * vec.x + car.right * vec.y + car.up * vec.z + car.location
 // }
@@ -298,7 +335,7 @@ pub fn curvature(v: f32) -> f32 {
         return 0.0018 - 4e-7 * v;
     }
 
-    0.
+    panic!("Invalid input velocity");
 }
 
 const VEC3_DOWN: Vec3 = Vec3 {
@@ -376,73 +413,54 @@ pub fn correct_for_posts(ball_location: Vec3, ball_radius: f32, target_left: Vec
     }
 }
 
+#[derive(Clone)]
 pub struct CarFieldRect {
-    u_1: f32,
-    u_2: f32,
-    v_1: f32,
-    v_2: f32,
-    a_1: f32,
-    b_1: f32,
-    a_2: f32,
-    b_2: f32,
+    goal_x: f32,
+    goal_y: f32,
+    field_y: f32,
+    field_x: f32,
+}
+
+impl Default for CarFieldRect {
+    fn default() -> Self {
+        Self {
+            goal_x: 0.,
+            goal_y: 0.,
+            field_y: 0.,
+            field_x: 0.,
+        }
+    }
 }
 
-// https://math.stackexchange.com/a/1938581/689600
 impl CarFieldRect {
-    pub fn from(car_hitbox: &Hitbox, car_y: f32) -> Self {
-        let margin = car_hitbox.length + car_hitbox.width / 2.;
+    pub fn from(car_hitbox: &Hitbox) -> Self {
+        let half_car_len = car_hitbox.length / 2.;
 
-        let mut length = 5120. - margin;
-        let extend = car_y.abs() >= length;
-        if extend {
-            length = 6000. - margin;
-        }
-
-        let width = 4093. - margin;
-
-        let field_1 = Vec3::new(length, width, 0.);
-        let field_2 = Vec3::new(-length, width, 0.);
-        let field_4 = Vec3::new(length, -width, 0.);
-
-        let u_1 = field_2.x - field_1.x;
-        let v_1 = field_2.y - field_1.y;
-        let u_2 = field_4.x - field_1.x;
-        let v_2 = field_4.y - field_1.y;
-
-        CarFieldRect {
-            u_1,
-            v_1,
-            u_2,
-            v_2,
-            a_1: field_1.x * u_1 + field_1.y * v_1,
-            b_1: field_2.x * u_1 + field_2.y * v_1,
-            a_2: field_1.x * u_2 + field_1.y * v_2,
-            b_2: field_4.x * u_2 + field_4.y * v_2,
+        Self {
+            goal_x: 893. - car_hitbox.width,
+            goal_y: 6000. - car_hitbox.length,
+            field_y: 5120. - half_car_len,
+            field_x: 4093. - half_car_len,
         }
     }
 
-    pub fn is_triangle_in(&self, p: [[f32; 3]; 3]) -> bool {
-        self.is_point_in(p[0]) && self.is_point_in(p[1]) && self.is_point_in(p[2])
+    // pub fn are_points_in(&self, s: &[[f32; 3]]) -> bool {
+    //     for p in s {
+    //         if !self.is_point_in(p) {
+    //             return false;
+    //         }
+    //     }
+
+    //     return true;
+    // }
+
+    pub fn is_point_in(&self, p: &[f32; 3]) -> bool {
+        if p[0].abs() > self.goal_x {
+            return p[0].abs() < self.field_x && p[1].abs() < self.field_y;
+        }
+
+        return p[1].abs() < self.goal_y;
     }
-
-    fn is_point_in(&self, p: [f32; 3]) -> bool {
-        let p_1 = p[0] * self.u_1 + p[1] * self.v_1;
-        let p_2 = p[0] * self.u_2 + p[1] * self.v_2;
-
-        self.a_1 < p_1 && p_1 < self.b_1 && self.a_2 < p_2 && p_2 < self.b_2
-    }
-}
-
-fn path_section_endpoints(path: &DubinsPath) -> Result<[[f32; 3]; 3], DubinsError> {
-    let dists = [path.param[0] * path.rho, (path.param[0] + path.param[1]) * path.rho, (path.param[0] + path.param[1] + path.param[2]) * path.rho];
-
-    Ok([path_sample(path, dists[0])?, path_sample(path, dists[1])?, path_sample(path, dists[2] - f32::EPSILON)?])
-}
-
-fn path_section_midpoints(path: &DubinsPath) -> Result<[[f32; 3]; 3], DubinsError> {
-    let dists = [path.param[0] * path.rho, (path.param[0] + path.param[1]) * path.rho, (path.param[0] + path.param[1] + path.param[2]) * path.rho];
-
-    Ok([path_sample(path, dists[0] / 2.)?, path_sample(path, dists[1] / 2.)?, path_sample(path, dists[2] / 2.)?])
 }
 
 fn path_endpoint_to_vec3(endpoint: [f32; 3]) -> Vec3 {
@@ -458,7 +476,64 @@ fn radius_from_local_point(a: Vec3) -> f32 {
     1. / (2. * b.y / a.dot(&b)).abs()
 }
 
-pub fn analyze_target(ball: &Box<Ball>, car: &Car, shot_vector: Vec3, time_remaining: f32, get_target: bool, validate: bool) -> Result<([f32; 4], Option<Vec3>, Option<DubinsPath>), DubinsError> {
+fn shortest_path_in_validate(q0: [f32; 3], q1: [f32; 3], rho: f32, car_field: &CarFieldRect) -> Result<(DubinsPath, [[f32; 3]; 3]), DubinsError> {
+    let mut best_cost = INFINITY;
+    let mut best_path = None;
+    let mut best_parts = None;
+
+    let intermediate_results = intermediate_results(q0, q1, rho)?;
+
+    for path_type in DubinsPathType::ALL {
+        if let Ok(param) = word(&intermediate_results, path_type) {
+            let cost = param[0] + param[1] + param[2];
+            if cost < best_cost {
+                let path = DubinsPath {
+                    qi: q0,
+                    rho,
+                    param,
+                    type_: path_type,
+                };
+
+                let mut valid = true;
+
+                for dist in [path.param[0] * path.rho / 2., (path.param[0] + path.param[1]) * path.rho / 2., (path.param[0] + path.param[1] + path.param[2]) * path.rho / 2.] {
+                    if !car_field.is_point_in(&path_sample(&path, dist)?) {
+                        valid = false;
+                        break;
+                    }
+                }
+
+                if !valid {
+                    continue;
+                }
+
+                let mut end_parts = [[0.; 3]; 3];
+
+                for (i, dist) in [path.param[0] * path.rho, (path.param[0] + path.param[1]) * path.rho, (path.param[0] + path.param[1] + path.param[2]) * path.rho].iter().enumerate() {
+                    end_parts[i] = path_sample(&path, *dist)?;
+
+                    if !car_field.is_point_in(&end_parts[i]) {
+                        valid = false;
+                        break;
+                    }
+                }
+
+                if valid {
+                    best_cost = cost;
+                    best_path = Some(path);
+                    best_parts = Some(end_parts);
+                }
+            }
+        }
+    }
+
+    match best_path {
+        Some(path) => Ok((path, best_parts.unwrap())),
+        None => Err(DubinsError::NoPath),
+    }
+}
+
+pub fn analyze_target(ball: &Ball, car: &Car, shot_vector: Vec3, time_remaining: f32, get_target: bool, validate: bool) -> Result<([f32; 4], Option<Vec3>, bool, Option<DubinsPath>), DubinsError> {
     let offset_target = ball.location - (shot_vector * ball.radius);
     let car_front_length = (car.hitbox_offset.x + car.hitbox.length) / 2.;
 
@@ -467,52 +542,68 @@ pub fn analyze_target(ball: &Box<Ball>, car: &Car, shot_vector: Vec3, time_remai
     let offset_distance = 640.;
     let exit_turn_point = offset_target - (shot_vector * offset_distance);
 
-    // lock onto the offset target if we're facing it
-    let local_offset = localize_2d(car, offset_target);
-    if local_offset.x > 0. {
-        let angle_to_shot = (ball.location - car.location).normalize().angle_2d(&shot_vector);
-        if angle_to_shot < 0.25 {
-            if !get_target {
-                return Ok(([0., 0., 0., end_distance], None, None));
-            }
+    let local_offset = localize_2d_location(car, offset_target);
 
-            let final_target = if local_offset.y.abs() < ball.radius + car.hitbox.width / 2. {
-                end_distance += local_offset.x;
+    if local_offset.x > 0. {
+        if local_offset.y.abs() < car.hitbox.width / 2. && car.forward.angle_2d(&shot_vector) < 0.15 {
+            end_distance += local_offset.x;
+
+            let final_target = if get_target {
                 Some(car.location + shot_vector * local_offset.x)
             } else {
-                end_distance += offset_distance + radius_from_local_point(localize_2d(car, exit_turn_point)) * angle_to_shot;
-                Some(exit_turn_point)
+                None
             };
+            return Ok(([0., 0., 0., end_distance], final_target, false, None));
+        } else {
+            let angle_to_shot = (ball.location - car.location).normalize().angle_2d(&shot_vector);
 
-            return Ok(([0., 0., 0., end_distance], final_target, None));
+            if angle_to_shot < 0.75 {
+                if car.forward.dot(&(exit_turn_point - car.location)) > 0. {
+                    // Use circle tangents and trig to calculate the turn radius
+                    let half_angle = shot_vector.angle_2d(&-car.forward) / 2.;
+                    let turn_rad = (car.location.dist_2d(exit_turn_point) / 2.) / half_angle.sin() * half_angle.tan();
+
+                    if turn_radius(car.local_velocity.x.abs()) < turn_rad && turn_rad < car.msmtr * 1.3 {
+                        end_distance += offset_distance + turn_rad * angle_to_shot;
+
+                        let final_target = if get_target {
+                            Some(exit_turn_point)
+                        } else {
+                            None
+                        };
+
+                        return Ok(([0., 0., 0., end_distance], final_target, true, None));
+                    }
+                } else if angle_to_shot < 0.25 && car.forward.angle_2d(&shot_vector) < 0.5 {
+                    end_distance += offset_distance + radius_from_local_point(local_offset) * angle_to_shot;
+
+                    let final_target = if get_target {
+                        Some(offset_target)
+                    } else {
+                        None
+                    };
+
+                    return Ok(([0., 0., 0., end_distance], final_target, false, None));
+                }
+            }
         }
     }
 
     end_distance += offset_distance;
 
-    if validate && end_distance + car.location.dist_2d(exit_turn_point) / time_remaining > car.max_speed {
+    if validate && (end_distance + car.location.dist_2d(exit_turn_point)) / time_remaining > car.max_speed {
         return Err(DubinsError::NoPath);
     }
 
     let q0 = [car.location.x, car.location.y, car.yaw];
     let q1 = [exit_turn_point.x, exit_turn_point.y, shot_vector.y.atan2(shot_vector.x)];
 
-    let path = shortest_path(q0, q1, car.msmtr * 1.1)?;
-    let end_parts = path_section_endpoints(&path)?;
-
-    if validate {
-        let car_field = CarFieldRect::from(&car.hitbox, car.location.y);
-        let midpoints = path_section_midpoints(&path)?;
-
-        if !car_field.is_triangle_in([path.qi, midpoints[0], end_parts[0]]) || !car_field.is_triangle_in([end_parts[1], midpoints[2], end_parts[2]]) {
-            return Err(DubinsError::NoPath);
-        }
-    }
+    let (path, end_parts) = shortest_path_in_validate(q0, q1, car.msmtr * 1.05, &car.field)?;
 
     let distances = [path.param[0] * path.rho, path.param[1] * path.rho, path.param[2] * path.rho];
 
     if !get_target {
-        return Ok(([distances[0], distances[1], distances[2], end_distance], None, Some(path)));
+        return Ok(([distances[0], distances[1], distances[2], end_distance], None, false, Some(path)));
     }
 
     let final_target = Some(path_endpoint_to_vec3(
@@ -527,7 +618,7 @@ pub fn analyze_target(ball: &Box<Ball>, car: &Car, shot_vector: Vec3, time_remai
         }],
     ));
 
-    Ok(([distances[0], distances[1], distances[2], end_distance], final_target, Some(path)))
+    Ok(([distances[0], distances[1], distances[2], end_distance], final_target, false, Some(path)))
 }
 
 fn throttle_acceleration(forward_velocity: f32) -> f32 {
@@ -549,7 +640,7 @@ pub fn can_reach_target(car: &Car, max_time: f32, distance_remaining: f32, is_fo
     let mut d = distance_remaining;
     let mut t_r = max_time;
     let mut b = car.boost as f32;
-    let mut v = car.velocity.dot(&car.forward);
+    let mut v = car.local_velocity.x;
 
     let direction = is_forwards as u8 as f32;
 
@@ -560,7 +651,7 @@ pub fn can_reach_target(car: &Car, max_time: f32, distance_remaining: f32, is_fo
 
         let r = d * direction / t_r;
 
-        if MIN_SPEED > r || r > car.max_speed || t_r < -SIMULATION_DT {
+        if t_r < -SIMULATION_DT || (is_forwards && r > car.max_speed) || (!is_forwards && MIN_SPEED > r) {
             return Err(());
         }
 
@@ -591,7 +682,7 @@ pub fn can_reach_target(car: &Car, max_time: f32, distance_remaining: f32, is_fo
         if throttle.signum() == v.signum() {
             v += throttle_accel * SIMULATION_DT * throttle;
         } else {
-            v += BRAKE_ACC_DT.copysign(-throttle);
+            v += BRAKE_ACC_DT.copysign(throttle);
         }
 
         if boost {
